/**
  Tiny OpenGL/GLUT particle system example program
  
  This version:
  	- Uses second-order dynamics
	- Treats "Z" as up (the right choice!)
	- Allows nice camera rotation using three vectors
	- Allows camera motion with "WASD" keys
	- Draws particles into a little ground texture, updated with a framebuffer object
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2009-01-21 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include "osl/vec4.h" /* 3D vector class */
#include "GL/glui.h" /* user interface widgets: http://glui.sourceforge.net/ */

#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z /* convert a vec3 to three float arguments */
#include <vector> /* std::vector container */


/* These variables are changed directly by the user interface code below */

/* Enable for anisotropic filtering */
int texaniso=1;

/* Mipmap level to mark with red, or -1 if none */
int red_level=-1;

/* Texture pixel format on graphics card */
int texformat=GL_RGBA8; /* updated by listbox */
const char *texformat_names[]={
	"GL_R3_G3_B2","GL_RGBA4","GL_RGBA8","GL_RGB10_A2","GL_RGB16",
	"GL_RGBA_FLOAT16","GL_RGBA_FLOAT32",
	0
};
GLenum texformat_enums[]={GL_R3_G3_B2, GL_RGBA4,GL_RGBA8,GL_RGB10_A2,GL_RGB16,
	GL_RGBA_FLOAT16_ATI,GL_RGBA_FLOAT32_ATI
};

/* Texture size, in pixels */
int texsize=32; /* updated by listbox */
const char *texsize_names[]={
	"2x2","4x4","19x19","32x32","256x256",
	0
};
GLenum texsize_enums[]={2,4,19,32,256};

/* Texture "wrap mode" in X and Y: */
int wrap_mode=GL_REPEAT;
const char *wrap_mode_names[]={
	"GL_REPEAT",
	"GL_CLAMP_TO_EDGE",
	"GL_CLAMP_TO_BORDER",
	"GL_CLAMP",
	0
};
GLenum wrap_mode_enums[]={
	GL_REPEAT,
	GL_CLAMP_TO_EDGE,
	GL_CLAMP_TO_BORDER,
	GL_CLAMP
};


/* Texture magnification mode/minification mode */
int minfilter_mode=GL_LINEAR_MIPMAP_LINEAR; /* updated by listbox */
int magfilter_mode=GL_LINEAR; /* updated by listbox */
const char *filtermode_names[]={
	"GL_NEAREST",
	"GL_LINEAR",
	"GL_LINEAR_MIPMAP_LINEAR",
	"GL_LINEAR_MIPMAP_NEAREST",
	"GL_NEAREST_MIPMAP_LINEAR",
	"GL_NEAREST_MIPMAP_NEAREST",
	0
};
GLenum filtermode_enums[]={
	GL_NEAREST,
	GL_LINEAR,
	GL_LINEAR_MIPMAP_LINEAR,
	GL_LINEAR_MIPMAP_NEAREST,
	GL_NEAREST_MIPMAP_LINEAR,
	GL_NEAREST_MIPMAP_NEAREST
};

/* Make a GLUI_Listbox with these modes */
void make_list(GLUI *glui,const char *title,int *mode_ptr,
	const char **names,GLenum *enums)
{
	GLUI_Listbox *list = new GLUI_Listbox( glui, title, mode_ptr);
	for (int i=0;names[i];i++) 
		list->add_item(enums[i],names[i]);
	list->do_selection(*mode_ptr);
}

/* Create the whole user interface.  Called from main */
void make_ui(GLUI *glui)
{
	make_list(glui,"size:",&texsize,texsize_names,texsize_enums);
	make_list(glui,"format:",&texformat,texformat_names,texformat_enums);
	make_list(glui,"wrap:",&wrap_mode,wrap_mode_names,wrap_mode_enums);
	//make_list(glui,"y wrap:",&wrap_y_mode,wrap_mode_names);
	make_list(glui,"min filter:",&minfilter_mode,
		filtermode_names,filtermode_enums);
	make_list(glui,"mag filter:",&magfilter_mode,
		filtermode_names,filtermode_enums);
	new GLUI_Checkbox( glui, "Anisotropic Filtering", &texaniso);
	(new GLUI_Spinner( glui, "Red Mipmap Level", &red_level))
		-> set_int_limits(-1,10);
}



/* Random Float: Return a random number between 0 and 1 */
float rf(void) { return (rand()&0xfff)*(1.0/0xfff); }
/* Random Vector: return a random vector in 3D, all axes between 0 and 1 */
vec3 rv(void) {return vec3(rf(),rf(),rf());}

bool key_down[256]; /* if true, the corresponding key is down */

/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}
/* Check if an error has happened--if so, exit */
void oglCheck(const char *where) {
	int err=0;
	if (0!=(err=glGetError())) {
		fprintf(stderr,"OpenGL error %d at %s\n",err,where);
		exit(1);
	}
}

/* 3D orientation */
/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,0,1), z(0,-1,0) {} /* point down Y axis */
	
	/* Reorient this coordinate frame by this far along in the X and Y axes.  
	   "dx" is the distance along the z axis to push the x axis; 
	   "dy" the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct an orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};
ortho_frame orient; /* camera orientation */
vec3 camera(0,-1.0,0.5); /* (initial) location of camera */

/* Make an OpenGL texture with our hardcoded properties */
GLuint make_texture(void)
{
	GLuint tex=0;
	glGenTextures(1,&tex);
	GLenum t=GL_TEXTURE_2D;
	glBindTexture(t,tex);
	
	int w=texsize, h=texsize;
	float *mem=new float[w*h*4];
	srand(1);
	for (int y=0;y<h;y++)
	for (int x=0;x<w;x++) 
	for (int c=0;c<4;c++) {
		float v=3.0*rf()-1.0;	/* from -1.0 to +2.0 */
		if (c==3) v=1.0; /* alpha is always 1.0 (opaque) */
		mem[c+4*(x+w*y)]=v;
	}
	
	//glTexImage2D(t,0,texformat, w,h,0, GL_RGBA,GL_UNSIGNED_BYTE,mem);
	gluBuild2DMipmaps(t,texformat, w,h, GL_RGBA,GL_FLOAT,mem);
	
	if (red_level>-1) { /* overwrite one mipmap level with red */
		w=w>>red_level; h=h>>red_level;
		if (w>0 && h>0) {
			for (int y=0;y<h;y++)
			for (int x=0;x<w;x++) 
			for (int c=0;c<4;c++) {
				float v=0.0;
				if (c==0 || c==3) v=1.0; /* 1 in red and alpha */
				mem[c+4*(x+w*y)]=v;
			}
			glTexImage2D(t,red_level,texformat, w,h,0, GL_RGBA,GL_FLOAT,mem);
		}
	}
	
	delete[] mem;
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,magfilter_mode); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,minfilter_mode);
	GLenum texWrap=wrap_mode;
	glTexParameteri(t, GL_TEXTURE_WRAP_S, texWrap); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, texWrap);
	
	glTexParameteri(t, GL_TEXTURE_MAX_ANISOTROPY_EXT,texaniso?16:1);
	
	glBindTexture(t,0); /* back to default texture */
	oglCheck("after make_texture");
	return tex;
}

void display(void) /* Draw everything in our window */
{
	oglCheck("before display function");
	static double elapsed=0.0; /* last elapsed display time (seconds) */
	static float dt=0.01, display_dt=0.0;
	static float sim_time=0.0;
	float big=300; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glFinish(); double start_t=oglTimer();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	GLuint dots_tex=make_texture();

/* Main camera setup */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.1,1000.0 /* Z range */
	);
	
	/* Physics and display timesteps: */
	float vel=0.5*display_dt; /* meters camera motion per frame */
	if (key_down[(int)'w']) camera-=vel*orient.z;
	if (key_down[(int)'s']) camera+=vel*orient.z;
	if (key_down[(int)'a']) camera-=vel*orient.x;
	if (key_down[(int)'d']) camera+=vel*orient.x;
	
	/* Don't allow head tilting! */
	orient.x.z=0.0; 
	orient.y=cross(orient.z,orient.x);
	orient.orthonormalize();
	
	vec3 target=camera-orient.z; /* camera looks down *negative* Z */
	gluLookAt(VEC3_TO_XYZ(camera), 
		VEC3_TO_XYZ(target), 
		VEC3_TO_XYZ(orient.y)
	);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Draw in a ground plane
	glBindTexture(GL_TEXTURE_2D,dots_tex);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	
	glBegin(GL_QUADS);
	glColor4f(1,1,1,1.0); /* light, opaque brown */
	glTexCoord2f(-big,-big); glVertex3f(-big,-big,floor);
	glTexCoord2f(+big,-big); glVertex3f(+big,-big,floor);
	glTexCoord2f(+big,+big); glVertex3f(+big,+big,floor);
	glTexCoord2f(-big,+big); glVertex3f(-big,+big,floor);
	glEnd();
	glDisable(GL_TEXTURE_2D);
	
	glDeleteTextures(1,&dots_tex);

	if (false) {
		// Draw in a 1-meter grid on the ground
		glEnable(GL_LINE_SMOOTH);
		glBegin(GL_LINES);
		glColor4f(0.0,0.0,0.0,0.1); /* transparent black -> grey */
		for (int v=-big;v<=big;v++) {
			glVertex3f(-big,v,floor); /* X lines */
			glVertex3f(+big,v,floor);
			glVertex3f(v,-big,floor); /* Y lines */
			glVertex3f(v,+big,floor);
		}
		glEnd();
	}
	
/* Draw axes */
	for (int axis=0;axis<3;axis++) {
		vec3 color(0.0);
		color[axis]=1.0;
		glColor3fv(color);
		
		glBegin(GL_LINES);
		glVertex3f(0,0,0);
		vec3 end(0.0); end[axis]=1.0;
		glVertex3fv(end);
		glEnd();
		
		glRasterPos3fv(end);
		glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,'X'+axis);
	}
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed+0.001;
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
	oglCheck("after display function");
}

int mouse_x=0,mouse_y=0;
void mouse(int button,int state,int x,int y) 
{ /* mouse being pressed or released--save position for motion */
	mouse_x=x; mouse_y=y;
}
void motion(int x, int y) { 
	float scale=0.01; /* radians rotation/frame per mouse pixel motion */
	orient.nudge(
		-scale*(x-mouse_x), 
		scale*(y-mouse_y)
	);
	mouse_x=x; mouse_y=y; /* save old mouse positions */
}
void keyboard(unsigned char key,int x,int y) {
	key_down[key]=true;
}
void keyboardUp(unsigned char key,int x,int y) {
	key_down[key]=false;
}

int main(int argc,char **argv) {
	glutInit(&argc,argv);  /* set up GLUT */
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	int win=glutCreateWindow("Texture Mode Demo (try WASD keys!)"); /* make a window */
	glewInit(); /* only AFTER create window! */
	
	// Make some popups & sliders (this is the only GLUI code you'll need!)
	GLUI *glui = GLUI_Master.create_glui_subwindow(win, 
		GLUI_SUBWINDOW_TOP);
	make_ui(glui);
	
	glutDisplayFunc(display); /* here's how to draw the window */
	glutMouseFunc(mouse);
	glutMotionFunc(motion); 
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardUp);
	glutMainLoop(); /* start drawing */
}
